package usherbrooke.ift630;

import java.io.IOException;
import java.io.BufferedReader;
import java.io.InputStreamReader;

import java.net.Socket;
import java.net.ServerSocket;

import java.util.ArrayList;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.Executors;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.LinkedBlockingQueue;

class Central {
	private ExecutorService threadsBus;
	private ExecutorService threadsSpareBusses;
	private ExecutorService threadsStop;
	private ExecutorService threadsPassenger;
	private ExecutorService threadsSocket;
	private ArrayList<Stop> stops;
	private ArrayList<Bus> busses;
	private ArrayList<Line> lines;
	private ArrayList<Passenger> passengers;
	private BlockingQueue<BusInformationMessage> blockingQueue;
	private ServerSocket socket;
	private Future<Void> socketFuture;
	private int socketPort;
	private boolean socketStop = false;

	private int timeEmbark;
	private int timeBetweenStops;
	private int numPassengersPerBus;

	// idk what to do with those
	private ArrayList<Integer> spacePerStop;
	private ArrayList<Integer> spacePerLine;

	private void sendBus(String msg) {
		Bus reqB = busses.get(Integer.parseInt(msg));
		Line l = reqB.getLine();
		// if >= 2 bus are full on that line, reset counter and add a bus
		if (spacePerLine.get(l.getId()) >= 2) {
			Bus b = new Bus(l, busses.size(), timeBetweenStops, timeEmbark, numPassengersPerBus);
			b.setSocketPort(socketPort);
			b.setBlockingQueue(blockingQueue);

			busses.add(b);

			threadsSpareBusses.submit(b);

			Logger.getInstance().print(0, "\t[CENTRAL] Sent a new bus on line " + l.getName());
		}
	}

	// should be used to see what lines need more bus and act accordingly.
	// This is hard to do w/ our implementation, since we cannot see what passenger
	// will embark on what line trivialy.
	// Therefore, when a bus is full, a new one will go get remaining passengers.
	private void refreshStops(String stop, String number) {
		int s = Integer.parseInt(stop);
		int n = Integer.parseInt(number);
		int l = 0;
		for (int k = 0 ; k < lines.size() ; k ++) {
			if (lines.get(k).getStops().contains(stops.get(s))) {
				l = k;
			}
		}
		
		spacePerStop.set(s, n);
		spacePerLine.set(l, spacePerLine.get(l) + 1);
		Logger.getInstance().print(0, "\t[CENTRAL] refreshed stop value for " + stops.get(s).getStopName());
	}

	// run TCP socket until an exception occurs
	private void runSocket() {
		if (socketFuture == null)
			// use a thread, so we dont block here
			threadsSocket.submit(() -> {
				try {
					while (!socketStop) {
						// blocks until client connects
						String txt;
						Socket s = socket.accept();

						BufferedReader rx = new BufferedReader(new InputStreamReader(s.getInputStream()));
						txt = rx.readLine();

						// split on ':'
						// 0 => who
						// 1 => id
						// 2 => arg if necessary
						String[] msg = txt.split(":");

						if (msg[0].startsWith("FULL"))
							sendBus(msg[1]);
						else if (msg[0].startsWith("STOP")) {
							refreshStops(msg[1], msg[2]);
						}

						// loop back to accept 
						s.close();
					}
				} catch (Exception e) {Logger.getInstance().print(1, "[CENTRAL] exc " + e.getMessage()); e.printStackTrace();}
			});
	}

	Central (int numThreadsBus, int numThreadsStop, int socketPort) throws IOException{
		threadsBus = Executors.newFixedThreadPool(numThreadsBus);
		threadsSpareBusses = Executors.newFixedThreadPool(numThreadsBus);
		threadsStop = Executors.newFixedThreadPool(numThreadsStop);
		threadsPassenger = Executors.newSingleThreadExecutor(); // doesnt need more! future balances themselves.
		threadsSocket = Executors.newSingleThreadExecutor();
		stops = new ArrayList<Stop>();
		busses = new ArrayList<Bus>();
		lines = new ArrayList<Line>();
		passengers = new ArrayList<Passenger>();
		spacePerStop = new ArrayList<Integer>();
		spacePerLine = new ArrayList<Integer>();
		blockingQueue = new LinkedBlockingQueue<BusInformationMessage>(); 

		this.socketPort = socketPort;
		socket = new ServerSocket(socketPort);
		runSocket();
	}

	// init stops
	public void createStops(int numStops, int numPassengersPerStop) {
		for (int k = 0; k < numStops; k++) {
			Stop s = new Stop(k, numPassengersPerStop);
			s.setBlockingQueue(blockingQueue);
			s.setSocketPort(socketPort);

			stops.add(s);
			s.setThreadPool(threadsStop);
			threadsStop.submit(s);
		}

		for (int k = 0 ; k < stops.size() ; k++)
			spacePerStop.add(k, 0);
	}

	// init lines (need stops to be init before)
	public void createLines(int numLines, int numStopPerBus, int numStops) throws Exception {
		for (int k = 0 ; k < numLines ; k++) {
			Stop s;
			if (numStopPerBus > stops.size()) 
				throw new Exception("not enough stops");
			ArrayList<Stop> s_list = new ArrayList<Stop>();
			while (s_list.size() < numStopPerBus) {
				do {
					s = stops.get((int) (Math.random() * numStops));
				} while (s_list.contains(s));
				s_list.add(s);
			}

			Line l = new Line(k, s_list);
			lines.add(l);
		}

		for (int k = 0 ; k < lines.size() ; k++)
			spacePerLine.add(k, 0);
	}

	// init busses
	public void createBusses(int numBusses, int timeBetweenStops, int timeEmbark, int numPassengersPerBus) throws IOException{
		this.numPassengersPerBus = numPassengersPerBus;
		this.timeBetweenStops = timeBetweenStops;
		for (int k = 0; k < numBusses; k++) {
			Line l = lines.get(k % lines.size());
			Bus b = new Bus(l, k, timeBetweenStops, timeEmbark, numPassengersPerBus);
			b.setSocketPort(socketPort);
			b.setBlockingQueue(blockingQueue);

			busses.add(b);
		}
	}

	// init pasengers
	public void createPassengers(int numPassengersPerStop, int numStopPerBus) {
	// make passenger
		int idPassenger = 0;
		for (Bus b : busses) {
			for (Stop start : b.getStops()) {
				// if not found or is terminus, dont use
				int idx = b.getStops().indexOf(start);
				if (idx == -1 || start == b.getTerminus())
					continue;

				// rand number of passenger per stop
				for (int k = 0; k < (int) (Math.random() * numPassengersPerStop); k++) {
					Stop dest = null;
					// dest is random from list of stops of this line

					dest = b.getStops().get(idx + 1 + (int) Math.round(Math.random() * (numStopPerBus - 2 - idx)));

					// create them!
					Passenger p = new Passenger(idPassenger, start, dest);
					passengers.add(p);
					start.addPassenger(p);
					idPassenger++;

					threadsPassenger.submit(p);
				}
			}
		}
	}

	// pretty print!
	public void printDetails() {
		for (Bus b : busses) {
			b.printDetails();
		}
	}

	// start bus thread
	public void startBusThreads() {
		for (Bus b : busses)
			threadsBus.submit(b);
	}

	// shutdowns all threand & return.
	public void shutdown() {
		// block stop and passenger pool once bus have finished
		try {
			threadsBus.shutdown();
			threadsBus.awaitTermination(Long.MAX_VALUE, TimeUnit.DAYS);
			threadsSpareBusses.shutdown();
			threadsSpareBusses.awaitTermination(Long.MAX_VALUE, TimeUnit.DAYS);
			threadsStop.shutdown();
			threadsStop.awaitTermination(Long.MAX_VALUE, TimeUnit.DAYS);
			threadsPassenger.shutdown();
			threadsPassenger.awaitTermination(Long.MAX_VALUE, TimeUnit.DAYS);
			socketStop = true;
			//threadsSocket.shutdown();
			//threadsSocket.awaitTermination(Long.MAX_VALUE, TimeUnit.DAYS);
		} catch (InterruptedException e) {}
	}
}
